@@ -4,5 +4,7 @@ from .midi_like import MIDILikeEncoding
 from .cp_word import CPWordEncoding
 from .mumidi import MuMIDIEncoding
 from .octuple import OctupleEncoding
+from .octuple_mono import OctupleMonoEncoding
 from .midi_tokenizer_base import quantize_note_times, detect_chords, merge_tracks, MIDITokenizer
-from .constants import MIDI_INSTRUMENTS, INSTRUMENT_CLASSES, INSTRUMENT_CLASSES_RANGES, CHORD_MAPS, DRUM_SETS
+from .constants import MIDI_INSTRUMENTS, INSTRUMENT_CLASSES, INSTRUMENT_CLASSES_RANGES, CHORD_MAPS, DRUM_SETS,\
+    CONTROL_CHANGES
