@@ -3,10 +3,11 @@ https://arxiv.org/abs/2101.02402
 
 """
 
+from pathlib import Path
 from typing import List, Tuple, Dict, Optional
 
 import numpy as np
-from miditoolkit import Instrument, Note
+from miditoolkit import MidiFile, Instrument, Note, TempoChange
 
 from .midi_tokenizer_base import MIDITokenizer, Event, detect_chords
 from .constants import *
@@ -55,6 +56,7 @@ class CPWordEncoding(MIDITokenizer):
             in the case of multitrack generation for instance
     :param params: can be a path to the parameter (json encoded) file or a dictionary
     """
+
     def __init__(self, pitch_range: range = PITCH_RANGE, beat_res: Dict[Tuple[int, int], int] = BEAT_RES,
                  nb_velocities: int = NB_VELOCITIES, additional_tokens: Dict[str, bool] = ADDITIONAL_TOKENS,
                  program_tokens: bool = PROGRAM_TOKENS, params=None):
@@ -76,34 +78,22 @@ class CPWordEncoding(MIDITokenizer):
             events.append(time_step)
         return events
 
-    def track_to_events(self, track: Instrument, time_division: int) -> List[List[Event]]:
+    def track_to_events(self, track: Instrument) -> List[List[Event]]:
         """ Converts a track (list of Note objects) into Event objects
 
         :param track: track object to convert
-        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
         :return: list of events
-                 the events should be in the order Bar -> Position -> Chord -> Pitch -> Velocity -> Duration
-        """
-        if self.additional_tokens['Ignore']:
-            return self._track_to_events_ignore(track, time_division)
-        return self._track_to_events_no_ignore(track, time_division)
-
-    def _track_to_events_ignore(self, track: Instrument, time_division: int) -> List[List[Event]]:
-        """ Converts a track (list of Note objects) into Event objects
-
-        :param track: track object to convert
-        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
-        :return: list of events
-                 the events should be in the order Bar -> Position -> Chord -> Pitch -> Velocity -> Duration
         """
         # Make sure the notes are sorted first by their onset (start) times, second by pitch
         # notes.sort(key=lambda x: (x.start, x.pitch))  # it should have been done in quantization function
-        ticks_per_frame = time_division // max(self.beat_res.values())
-        ticks_per_bar = time_division * 4
+        ticks_per_frame = self.current_midi_metadata['time_division'] // max(self.beat_res.values())
+        ticks_per_bar = self.current_midi_metadata['time_division'] * 4
         events = []  # list of lists of Event objects
 
         def create_event(time: int, bar=False, pos=None, pitch=None, vel=None, dur=None, chord=None, empty=False,
-                         text='') -> List[Event]:
+                         tempo=None, text='') -> List[Event]:
+            chordempty_idx = 0
+            temp_idx = -1
             event_template = [Event(name='Family', time=time, value='Metric', text=text),
                               Event(name='BarPosition', time=time, value='Ignore', text=text),
                               Event(name='Pitch', time=time, value='Ignore', text=text),
@@ -111,22 +101,28 @@ class CPWordEncoding(MIDITokenizer):
                               Event(name='Duration', time=time, value='Ignore', text=text)]
             if self.additional_tokens['Chord'] or self.additional_tokens['Empty']:
                 event_template += [Event(name='ChordEmpty', time=time, value='Ignore', text=text)]
+                chordempty_idx = -1
+            if self.additional_tokens['Tempo']:
+                event_template += [Event(name='Tempo', time=time, value='Ignore', text=text)]
+                if chordempty_idx == -1:
+                    chordempty_idx = -2
 
             if bar:
                 event_template[1] = Event(name='Bar', time=time, value=None, text=text)
                 if empty:
-                    event_template[5] = Event(name='Empty', time=time, value=None, text=text)
+                    event_template[chordempty_idx] = Event(name='Empty', time=time, value=None, text=text)
             elif pos is not None:
                 event_template[1] = Event(name='Position', time=time, value=pos, text=text)
+                if chord is not None:
+                    event_template[chordempty_idx] = Event(name='Chord', time=time, value=chord, text=text)
+                if tempo is not None:
+                    event_template[temp_idx] = Event(name='Tempo', time=time, value=tempo, text=text)
             elif pitch is not None:
                 event_template[0].value = 'Note'
                 event_template[2].value = pitch
                 event_template[3].value = vel
                 event_template[4].value = dur
 
-            if chord is not None:
-                event_template[5] = Event(name='Chord', time=time, value=chord, text=text)
-
             return event_template
 
         # Creates Bar and Position events
@@ -143,18 +139,37 @@ class CPWordEncoding(MIDITokenizer):
 
         # Creates the Pitch, Velocity and Duration events
         current_tick = -1
+        current_tempo_idx = 0
+        current_tempo = self.current_midi_metadata['tempo_changes'][current_tempo_idx].tempo
+        current_tempo = (np.abs(self.tempo_bins - current_tempo)).argmin()
         for note in track.notes:
             if note.pitch not in self.pitch_range:  # Notes to low or to high are discarded
                 continue
+            # Position
             if note.start != current_tick:
                 pos_index = int((note.start % ticks_per_bar) / ticks_per_frame)
-                events.append(create_event(int(note.start), pos=pos_index, text=str(note.start)))
+                if self.additional_tokens['Tempo']:
+                    # If the current tempo is not the last one
+                    if current_tempo_idx + 1 < len(self.current_midi_metadata['tempo_changes']):
+                        # Will loop over incoming tempo changes
+                        for tempo_change in self.current_midi_metadata['tempo_changes'][current_tempo_idx + 1:]:
+                            # If this tempo change happened before the current moment
+                            if tempo_change.time <= current_tick:
+                                current_tempo = (np.abs(self.tempo_bins - tempo_change.tempo)).argmin()
+                                current_tempo_idx += 1  # update tempo value (might not change) and index
+                            elif tempo_change.time > current_tick:
+                                break  # this tempo change is beyond the current time step, we break the loop
+                    events.append(create_event(int(note.start), pos=pos_index, tempo=current_tempo,
+                                               text=str(note.start)))
+                else:
+                    events.append(create_event(int(note.start), pos=pos_index, text=str(note.start)))
                 current_tick = note.start
 
             # Note
             velocity_index = (np.abs(self.velocity_bins - note.velocity)).argmin()
             duration = note.end - note.start
-            dur_index = np.argmin(np.abs([ticks - duration for ticks in self.durations_ticks[time_division]]))
+            dur_index = np.argmin(np.abs([ticks - duration for ticks in
+                                          self.durations_ticks[self.current_midi_metadata['time_division']]]))
             dur_value = '.'.join(map(str, self.durations[dur_index]))
             events.append(create_event(int(note.start), pitch=note.pitch, vel=velocity_index, dur=dur_value,
                                        text=f'{duration} ticks'))
@@ -163,7 +178,7 @@ class CPWordEncoding(MIDITokenizer):
 
         # Adds chord events if specified
         if self.additional_tokens['Chord'] and not track.is_drum:
-            chord_events = detect_chords(track.notes, time_division)
+            chord_events = detect_chords(track.notes, self.current_midi_metadata['time_division'])
             count = 0
             for chord_event in chord_events:
                 for e, compound_token in enumerate(events[count:]):
@@ -174,75 +189,41 @@ class CPWordEncoding(MIDITokenizer):
 
         return events
 
-    def _track_to_events_no_ignore(self, track: Instrument, time_division: int) -> \
-            List[List[Event]]:
-        """ Converts a track (list of Note objects) into Event objects
+    def tokens_to_midi(self, tokens: List[List[int]], programs: Optional[List[Tuple[int, bool]]] = None,
+                       output_path: Optional[str] = None, time_division: Optional[int] = TIME_DIVISION) -> MidiFile:
+        """ Override the parent class method
+        Convert multiple sequences of tokens into a multitrack MIDI and save it.
+        The tokens will be converted to event objects and then to a miditoolkit.MidiFile object.
+        NOTE: for multitrack with tempo, only the tempo tokens of the first
+            decoded track will be used for the MIDI
 
-        :param track: track object to convert
-        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
-        :return: list of events
-                 the events should be in the order Bar -> Position -> Chord -> Pitch -> Velocity -> Duration
+        :param tokens: list of lists of tokens to convert, each list inside the
+                       first list corresponds to a track
+        :param programs: programs of the tracks
+        :param output_path: path to save the file (with its name, e.g. music.mid),
+                        leave None to not save the file
+        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI to create)
+        :return: the midi object (miditoolkit.MidiFile)
         """
-        # Make sure the notes are sorted first by their onset (start) times, second by pitch
-        # notes.sort(key=lambda x: (x.start, x.pitch))  # it should have been done in quantization function
-        ticks_per_frame = time_division // max(self.beat_res.values())
-        ticks_per_bar = time_division * 4
-        events = []  # list of lists of Event objects
+        midi = MidiFile(ticks_per_beat=time_division)
+        for i, track_tokens in enumerate(tokens):
+            if programs is not None:
+                track, tempos = self.tokens_to_track(track_tokens, time_division, programs[i])
+            else:
+                track, tempos = self.tokens_to_track(track_tokens, time_division)
+            midi.instruments.append(track)
+            if i == 0:  # only keep tempo changes of the first track
+                midi.tempo_changes = tempos
+                midi.tempo_changes[0].time = 0
 
-        # Creates Bar and Position events
-        bar_ticks = np.arange(0, max(n.end for n in track.notes) + ticks_per_bar, ticks_per_bar)
-        for t, tick in enumerate(bar_ticks):  # creating a "Bar" event at each beginning of bars
-
-            events.append([Event(name='Family', time=tick, value='Metric', text=t),
-                           Event(name='Bar', time=tick, value=None, text=t)])
-
-            if self.additional_tokens['Empty']:
-                # We consider a note inside a bar when its onset time is within the bar
-                # as it is how the note messages will be put in the sequence
-                notes_in_this_bar = [note for note in track.notes if tick <= note.start < tick + ticks_per_bar]
-                if len(notes_in_this_bar) == 0:
-                    events[-1].append(Event(name='Empty', time=tick, value=None, text=t))
-
-        # Creates the Pitch, Velocity and Duration events
-        current_tick = -1
-        for note in track.notes:
-            if note.pitch not in self.pitch_range:  # Notes to low or to high are discarded
-                continue
-            if note.start != current_tick:
-                pos_index = int((note.start % ticks_per_bar) / ticks_per_frame)
-                events.append([Event(name='Family', time=note.start, value='Metric', text=note.start),
-                               Event(name='Position', time=note.start, value=pos_index, text=note.start)])
-
-                current_tick = note.start
-
-            # Note
-            velocity_index = (np.abs(self.velocity_bins - note.velocity)).argmin()
-            duration = note.end - note.start
-            dur_index = np.argmin(np.abs([ticks - duration for ticks in self.durations_ticks[time_division]]))
-            events.append([Event(name='Family', time=note.start, value='Note', text=note.start),
-                           Event(name='Pitch', time=note.start, value=note.pitch, text=note.pitch),
-                           Event(name='Velocity', time=note.start, value=velocity_index,
-                                 text=f'{note.velocity}/{self.velocity_bins[velocity_index]}'),
-                           Event(name='Duration', time=note.start, value='.'.join(map(str, self.durations[dur_index])),
-                                 text=f'{duration} ticks')])
-
-        events.sort(key=lambda x: x[0].time)
-
-        # Adds chord events if specified
-        if self.additional_tokens['Chord'] and not track.is_drum:
-            chord_events = detect_chords(track.notes, time_division)
-            count = 0
-            for chord_event in chord_events:
-                for e, compound_token in enumerate(events[count:]):
-                    if compound_token[1].time == chord_event.time and compound_token[1].name == 'Position':
-                        compound_token.append(chord_event)
-                        count = e
-                        break
-
-        return events
+        # Write MIDI file
+        if output_path:
+            Path(output_path).mkdir(parents=True, exist_ok=True)
+            midi.dump(output_path)
+        return midi
 
     def events_to_track(self, events: List[List[Event]], time_division: int,
-                        program: Optional[Tuple[int, bool]] = (0, False)) -> Instrument:
+                        program: Optional[Tuple[int, bool]] = (0, False)) -> Tuple[Instrument, List[TempoChange]]:
         """ Transform a list of Event objects into an instrument object
 
         :param events: list of Event objects to convert to a track
@@ -254,16 +235,19 @@ class CPWordEncoding(MIDITokenizer):
 
         name = 'Drums' if program[1] else MIDI_INSTRUMENTS[program[0]]['name']
         instrument = Instrument(program[0], is_drum=program[1], name=name)
+        if self.additional_tokens['Tempo']:
+            tempo_changes = [TempoChange(TEMPO, -1)]  # mock the first tempo change to optimize below
+        else:  # default
+            tempo_changes = [TempoChange(TEMPO, 0)] * 2  # the first will be deleted at the end of the method
         current_tick = 0
         current_bar = -1
-        offset = 1 if self.additional_tokens['Ignore'] else 0
 
         for compound_token in events:
             token_type = compound_token[0].value
             if token_type == 'Note':
-                pitch = int(compound_token[1 + offset].value)
-                vel = int(self.velocity_bins[int(compound_token[2 + offset].value)])
-                beat, pos, res = map(int, compound_token[3 + offset].value.split('.'))
+                pitch = int(compound_token[2].value)
+                vel = int(self.velocity_bins[int(compound_token[3].value)])
+                beat, pos, res = map(int, compound_token[4].value.split('.'))
                 duration = (beat * res + pos) * time_division // res
                 instrument.notes.append(Note(vel, pitch, current_tick, current_tick + duration))
             elif token_type == 'Metric':
@@ -272,8 +256,12 @@ class CPWordEncoding(MIDITokenizer):
                     current_tick = current_bar * time_division * 4
                 elif compound_token[1].name == 'Position':
                     current_tick = current_bar * time_division * 4 + int(compound_token[1].value) * ticks_per_frame
-
-        return instrument
+                    if self.additional_tokens['Tempo']:
+                        tempo = int(self.tempo_bins[int(compound_token[-1].value)])
+                        if tempo != tempo_changes[-1].tempo:
+                            tempo_changes.append(TempoChange(tempo, current_tick))
+        del tempo_changes[0]
+        return instrument, tempo_changes
 
     def create_vocabulary(self, program_tokens: bool) -> Tuple[dict, dict, dict]:
         """ Create the tokens <-> event dictionaries
@@ -287,42 +275,48 @@ class CPWordEncoding(MIDITokenizer):
         :param program_tokens: creates tokens for MIDI programs in the dictionary
         :return: the dictionaries, one for each translation
         """
-        event_to_token = {'PAD_None': 0, 'Bar_None': 1, 'Family_Note': 3, 'Family_Metric': 4}
+        event_to_token = {'PAD_None': 0, 'Bar_None': 1, 'Family_Note': 2, 'Family_Metric': 3}
         token_type_indices = {'Pad': [0], 'BarPosition': [1], 'Family': [2, 3]}
-        count = 5
-        if self.additional_tokens['Empty']:  # will be considered in the Metric family
-            event_to_token['Empty_None'] = count
-            token_type_indices['ChordEmpty'] = [count]
-            count += 1
+        count = 4
 
         # PITCH
-        token_type_indices['Pitch'] = list(range(count, count + len(self.pitch_range)))
+        event_to_token['Pitch_Ignore'] = count
+        token_type_indices['Pitch'] = list(range(count, count + len(self.pitch_range) + 1))
+        count += 1
         for i in self.pitch_range:
             event_to_token[f'Pitch_{i}'] = count
             count += 1
 
         # VELOCITY
-        token_type_indices['Velocity'] = list(range(count, count + len(self.velocity_bins)))
+        event_to_token['Velocity_Ignore'] = count
+        token_type_indices['Velocity'] = list(range(count, count + len(self.velocity_bins) + 1))
+        count += 1
         for i in range(len(self.velocity_bins)):
             event_to_token[f'Velocity_{i}'] = count
             count += 1
 
         # DURATION
-        token_type_indices['Duration'] = list(range(count, count + len(self.durations)))
+        event_to_token['Duration_Ignore'] = count
+        token_type_indices['Duration'] = list(range(count, count + len(self.durations) + 1))
+        count += 1
         for i in range(0, len(self.durations)):
             event_to_token[f'Duration_{".".join(map(str, self.durations[i]))}'] = count
             count += 1
 
         # POSITION
         nb_positions = max(self.beat_res.values()) * 4  # 4/4 time signature
-        token_type_indices['BarPosition'] += list(range(count, count + nb_positions))
+        event_to_token['BarPosition_Ignore'] = count
+        token_type_indices['BarPosition'] += list(range(count, count + nb_positions + 1))
+        count += 1
         for i in range(0, nb_positions):
             event_to_token[f'Position_{i}'] = count
             count += 1
 
         # CHORD
         if self.additional_tokens['Chord']:
-            token_type_indices['Chord'] = list(range(count, count + 3 + len(CHORD_MAPS)))
+            event_to_token['ChordEmpty_Ignore'] = count
+            token_type_indices['ChordEmpty'] = list(range(count, count + 3 + len(CHORD_MAPS) + 1))
+            count += 1
             for i in range(3, 6):  # non recognized chords, just considers the nb of notes (between 3 and 5 only)
                 event_to_token[f'Chord_{i}'] = count
                 count += 1
@@ -330,28 +324,26 @@ class CPWordEncoding(MIDITokenizer):
                 event_to_token[f'Chord_{chord_quality}'] = count
                 count += 1
 
-        # IGNORE
-        if self.additional_tokens['Ignore']:
-            event_to_token['BarPosition_Ignore'] = count
-            token_type_indices['BarPosition'] += [count]
-            event_to_token['Pitch_Ignore'] = count + 1
-            token_type_indices['Pitch'] += [count + 1]
-            event_to_token['Velocity_Ignore'] = count + 2
-            token_type_indices['Velocity'] += [count + 2]
-            event_to_token['Duration_Ignore'] = count + 3
-            token_type_indices['Duration'] += [count + 3]
-            count += 4
-            if self.additional_tokens['Chord'] or self.additional_tokens['Empty']:
+        # EMPTY
+        if self.additional_tokens['Empty']:
+            if 'ChordEmpty_Ignore' not in event_to_token:
                 event_to_token['ChordEmpty_Ignore'] = count
-                try:
-                    token_type_indices['ChordEmpty'] += [count]
-                except KeyError as _:
-                    token_type_indices['ChordEmpty'] = [count]
+                token_type_indices['ChordEmpty'] = [count, count+1]
                 count += 1
-
-            token_type_indices['Empty'] = [count]
+            else:
+                token_type_indices['ChordEmpty'] += [count]
+            event_to_token['Empty_None'] = count
             count += 1
 
+        # TEMPO
+        if self.additional_tokens['Tempo']:
+            event_to_token['Tempo_Ignore'] = count
+            token_type_indices['Tempo'] = list(range(count, count + len(self.tempo_bins) + 1))
+            count += 1
+            for i in range(len(self.tempo_bins)):
+                event_to_token[f'Tempo_{i}'] = count
+                count += 1
+
         # PROGRAM
         if program_tokens:
             event_to_token['Family_Program'] = count
