@@ -9,9 +9,10 @@ from pathlib import Path, PurePath
 from typing import List, Tuple, Dict, Optional, Union
 
 import numpy as np
-from miditoolkit import Instrument, Note, MidiFile
+from miditoolkit import MidiFile, Instrument, Note
 
-from .midi_tokenizer_base import MIDITokenizer, Event, detect_chords, quantize_note_times, remove_duplicated_notes
+from .midi_tokenizer_base import MIDITokenizer, Event, detect_chords, quantize_note_times, quantize_tempos, \
+    remove_duplicated_notes
 from .constants import *
 
 
@@ -40,7 +41,6 @@ class MuMIDIEncoding(MIDITokenizer):
                  beat_res: Dict[Tuple[int, int], int] = BEAT_RES, nb_velocities: int = NB_VELOCITIES,
                  additional_tokens: Dict[str, bool] = ADDITIONAL_TOKENS, program_tokens: bool = PROGRAM_TOKENS,
                  params=None):
-        additional_tokens['Ignore'] = False  # Incompatible additional tokens
         self.drum_pitch_range = drum_pitch_range
         # used in place of positional encoding
         self.max_bar_embedding = 60  # this attribute might increase during encoding
@@ -50,6 +50,7 @@ class MuMIDIEncoding(MIDITokenizer):
         """ Override the parent class method to include additional parameter drum pitch range
         Saves the base parameters of this encoding in a txt file
         Useful to keep track of how a dataset has been tokenized / encoded
+        It will also save the name of the class used, i.e. the encoding strategy
 
         :param out_dir: output directory to save the file
         """
@@ -58,7 +59,8 @@ class MuMIDIEncoding(MIDITokenizer):
             json.dump({'pitch_range': (self.pitch_range.start, self.pitch_range.stop),
                        'drum_pitch_range': (self.drum_pitch_range.start, self.drum_pitch_range.stop),
                        'beat_res': self.beat_res, 'nb_velocities': len(self.velocity_bins),
-                       'additional_tokens': self.additional_tokens, 'max_bar_embedding': self.max_bar_embedding},
+                       'additional_tokens': self.additional_tokens, 'encoding': self.__class__.__name__,
+                       'max_bar_embedding': self.max_bar_embedding},
                       outfile)
 
     def midi_to_tokens(self, midi: MidiFile) -> List[List[int]]:
@@ -76,6 +78,12 @@ class MuMIDIEncoding(MIDITokenizer):
             self.durations_ticks[midi.ticks_per_beat] = [(beat * res + pos) * midi.ticks_per_beat // res
                                                          for beat, pos, res in self.durations]
 
+        self.current_midi_metadata = {'time_division': midi.ticks_per_beat,
+                                      'tempo_changes': midi.tempo_changes,
+                                      'time_sig_changes': midi.time_signature_changes,
+                                      'key_sig_changes': midi.key_signature_changes}
+        quantize_tempos(midi.tempo_changes, midi.ticks_per_beat, max(self.beat_res.values()))
+
         ticks_per_frame = midi.ticks_per_beat // max(self.beat_res.values())
         ticks_per_bar = midi.ticks_per_beat * 4
 
@@ -85,7 +93,7 @@ class MuMIDIEncoding(MIDITokenizer):
             quantize_note_times(track.notes, midi.ticks_per_beat, max(self.beat_res.values()))  # adjusts notes timings
             track.notes.sort(key=lambda x: (x.start, x.pitch))  # sort notes
             remove_duplicated_notes(track.notes)  # remove possible duplicated notes
-            note_events += self.track_to_events(track, midi.ticks_per_beat)
+            note_events += self.track_to_events(track)
 
             # Check bar embedding limit, update if needed
             nb_bars = ceil(max(note.end for note in track.notes) / ticks_per_bar)
@@ -106,45 +114,79 @@ class MuMIDIEncoding(MIDITokenizer):
         current_bar = -1
         current_pos = -1
         current_track = -2  # doesnt exist
+        current_tempo_idx = 0
+        current_tempo = self.current_midi_metadata['tempo_changes'][current_tempo_idx].tempo
+        current_tempo = (np.abs(self.tempo_bins - current_tempo)).argmin()
         for note_event in note_events:
+            # (Tempo) update tempo values current_tempo
+            if self.additional_tokens['Tempo']:
+                # If the current tempo is not the last one
+                if current_tempo_idx + 1 < len(self.current_midi_metadata['tempo_changes']):
+                    # Will loop over incoming tempo changes
+                    for tempo_change in self.current_midi_metadata['tempo_changes'][current_tempo_idx + 1:]:
+                        # If this tempo change happened before the current moment
+                        if tempo_change.time <= current_tick:
+                            current_tempo = (np.abs(self.tempo_bins - tempo_change.tempo)).argmin()
+                            current_tempo_idx += 1  # update tempo value (might not change) and index
+                        elif tempo_change.time > current_tick:
+                            break  # this tempo change is beyond the current time step, we break the loop
             # Positions and bars
             if note_event[0].time != current_tick:
                 pos_index = int((note_event[0].time % ticks_per_bar) / ticks_per_frame)
                 current_tick = note_event[0].time
                 current_pos = pos_index
                 current_track = -2  # reset
+                # (New bar)
                 if current_bar < current_tick // ticks_per_bar:
                     nb_new_bars = current_tick // ticks_per_bar - current_bar
                     for i in range(nb_new_bars):
-                        tokens.append([self.event2token['Bar_None'],
-                                       self.event2token['Position_Ignore'],
-                                       self.event2token[f'Bar_{current_bar + i + 1}']])
+                        bar_token = [self.event2token['Bar_None'],
+                                     self.event2token['Position_Ignore'],
+                                     self.event2token[f'Bar_{current_bar + i + 1}']]
+                        if self.additional_tokens['Tempo']:
+                            bar_token.append(self.event2token[f'Tempo_{current_tempo}'])
+                        tokens.append(bar_token)
+                        # (Empty bar)
+                        if self.additional_tokens['Empty'] and i+1 != nb_new_bars and nb_new_bars > 1:
+                            empty_token = [self.event2token['Empty_None'],
+                                           self.event2token['Position_Ignore'],
+                                           self.event2token[f'Bar_{current_bar + i + 1}']]
+                            if self.additional_tokens['Tempo']:
+                                empty_token.append(self.event2token[f'Tempo_{current_tempo}'])
+                            tokens.append(empty_token)
                     current_bar += nb_new_bars
-                tokens.append([self.event2token['Position_None'],
-                               self.event2token[f'Position_{current_pos}'],
-                               self.event2token[f'Bar_{current_bar}']])
+                # Position
+                pos_token = [self.event2token['Position_None'],
+                             self.event2token[f'Position_{current_pos}'],
+                             self.event2token[f'Bar_{current_bar}']]
+                if self.additional_tokens['Tempo']:
+                    pos_token.append(self.event2token[f'Tempo_{current_tempo}'])
+                tokens.append(pos_token)
             # Tracks (programs)
             if note_event[0].text != current_track:
                 current_track = note_event[0].text
-                tokens.append([self.event2token[f'Program_{current_track}'],
+                track_token = [self.event2token[f'Program_{current_track}'],
                                self.event2token[f'Position_{current_pos}'],
-                               self.event2token[f'Bar_{current_bar}']])
+                               self.event2token[f'Bar_{current_bar}']]
+                if self.additional_tokens['Tempo']:
+                    track_token.append(self.event2token[f'Tempo_{current_tempo}'])
+                tokens.append(track_token)
 
             # Adding bar and position tokens to notes for positional encoding
             note_event[0] = self.event2token[f'{note_event[0].name}_{note_event[0].value}']
-            tokens.append(note_event + [self.event2token[f'Position_{current_pos}'],
-                                        self.event2token[f'Bar_{current_bar}']])
+            note_event += [self.event2token[f'Position_{current_pos}'], self.event2token[f'Bar_{current_bar}']]
+            if self.additional_tokens['Tempo']:
+                note_event.append(self.event2token[f'Tempo_{current_tempo}'])
+            tokens.append(note_event)
 
         return tokens
 
-    def track_to_events(self, track: Instrument, time_division: int) -> List[List[Event]]:
+    def track_to_events(self, track: Instrument) -> List[List[Event]]:
         """ Converts a track (list of Note objects) into Event objects
         This only create pitch, velocity and duration events
 
         :param track: track object to convert
-        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
         :return: list of events
-                 the events should be in the order Bar -> Position -> Chord -> Pitch -> Velocity -> Duration
         """
         # Make sure the notes are sorted first by their onset (start) times, second by pitch
         # notes.sort(key=lambda x: (x.start, x.pitch))  # done in midi_to_tokens
@@ -156,7 +198,8 @@ class MuMIDIEncoding(MIDITokenizer):
             # Note
             velocity_index = (np.abs(self.velocity_bins - note.velocity)).argmin()
             duration = note.end - note.start
-            dur_index = np.argmin(np.abs([ticks - duration for ticks in self.durations_ticks[time_division]]))
+            dur_index = np.argmin(np.abs([ticks - duration for ticks in
+                                          self.durations_ticks[self.current_midi_metadata['time_division']]]))
             if not track.is_drum:
                 events.append([Event(name='Pitch', time=note.start, value=note.pitch, text=track.program),
                                self.event2token[f'Velocity_{velocity_index}'],
@@ -168,7 +211,7 @@ class MuMIDIEncoding(MIDITokenizer):
 
         # Adds chord events if specified
         if self.additional_tokens['Chord'] and not track.is_drum:
-            chords = detect_chords(track.notes, time_division)
+            chords = detect_chords(track.notes, self.current_midi_metadata['time_division'])
             unsqueezed = []
             for c in range(len(chords)):
                 chords[c].text = track.program
@@ -182,6 +225,11 @@ class MuMIDIEncoding(MIDITokenizer):
         """ Override the parent class method
         Convert multiple sequences of tokens into a multitrack MIDI and save it.
         The tokens will be converted to event objects and then to a miditoolkit.MidiFile object.
+        A time step is a list of tokens where:
+        (list index: token type)
+        0: Pitch / Position / Bar
+        (1: Velocity)
+        (2: Duration)
 
         :param tokens: list of lists of tokens to convert, each list inside the
                        first list corresponds to a track
@@ -314,6 +362,13 @@ class MuMIDIEncoding(MIDITokenizer):
                 event_to_token[f'Chord_{chord_quality}'] = count
                 count += 1
 
+        # TEMPO
+        if self.additional_tokens['Tempo']:
+            token_type_indices['Tempo'] = list(range(count, count + len(self.tempo_bins)))
+            for i in range(len(self.tempo_bins)):
+                event_to_token[f'Tempo_{i}'] = count
+                count += 1
+
         # EMPTY
         if self.additional_tokens['Empty']:
             event_to_token['Empty_None'] = count
